﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.01
@EndOfDesignText@
'version 3.00 (beta)
Sub Class_Globals
	Private buffer() As Byte
	Public mWidth, mHeight As Int
	Private ai, ri, gi, bi As Int
	Type PremultipliedColor (a As Int, r As Int, g As Int, b As Int)
	Type ARGBColor (a As Int, r As Int, g As Int, b As Int)
	Private tempARGB As ARGBColor
	Private tempPM As PremultipliedColor
#if B4A
	Private b4abmp As Bitmap
#else if B4J
	Private WriteableImage As JavaObject
#end if
	Public TargetRect As B4XRect
	
End Sub

Public Sub Initialize (Width As Int, Height As Int)
	mWidth = Width
	mHeight = Height
	Dim buffer(mWidth * mHeight * 4) As Byte
	#if B4A
	b4abmp.InitializeMutable(mWidth, mHeight) 'ignore
	#End if
	
	#if B4A or B4i
	ai = 3
	ri = 0
	gi = 1
	bi = 2
	#else if B4J
	ai = 3
	ri = 2
	gi = 1
	bi = 0
	WriteableImage.InitializeNewInstance("javafx.scene.image.WritableImage", Array(Width, Height))
	#end if
	TargetRect.Initialize(0, 0, mWidth, mHeight)
End Sub

Public Sub CopyPixelsFromBitmap (SourceBitmap As B4XBitmap)
	If SourceBitmap.Width <> mWidth Or SourceBitmap.Height <> mHeight Then
		SourceBitmap = SourceBitmap.Resize(mWidth, mHeight, False)
	End If
	#if B4A
	Dim jo As JavaObject = SourceBitmap
	Dim format As Object = jo.RunMethod("getConfig", Null)
	If "ARGB_8888" <> format Then
		Log("Unsupported format: " & format)
		Return
	End If
	Dim bf As JavaObject
	bf = bf.InitializeStatic("java.nio.ByteBuffer").RunMethodJO("wrap", Array(buffer))
	jo.RunMethod("copyPixelsToBuffer", Array(bf))
	#else if B4J
	Dim jo As JavaObject = SourceBitmap
	Dim reader As JavaObject = jo.RunMethod("getPixelReader", Null)
	Dim PixelFormat As JavaObject
	PixelFormat.InitializeStatic("javafx.scene.image.PixelFormat")
	reader.RunMethod("getPixels", Array(0, 0, mWidth, mHeight, PixelFormat.RunMethod("getByteBgraPreInstance", Null), _
       buffer, 0, mWidth * 4))
	 #Else If B4i
	Dim no As NativeObject = Me
	no.RunMethod("UIImageToBuffer::::", Array(SourceBitmap, no.ArrayToNSData(buffer), mWidth, mHeight))
	#End If
End Sub

Public Sub SetColor(x As Int, y As Int, clr As Int)
	ColorToARGB(clr, tempARGB)
	SetARGB(x, y, tempARGB)
End Sub

Public Sub ColorToARGB (clr As Int, Result As ARGBColor) As ARGBColor
	Result.a = Bit.And(0xff, Bit.ShiftRight(clr, 24))
	Result.r = Bit.And(0xff, Bit.ShiftRight(clr, 16))
	Result.g =  Bit.And(0xff, Bit.ShiftRight(clr, 8))
	Result.b =  Bit.And(0xff, clr)
	Return Result
End Sub

Public Sub ARGBToPremultipliedColor (ARGB As ARGBColor, pm As PremultipliedColor) As PremultipliedColor
	Dim a As Float = ARGB.a / 255
	pm.a = ARGB.a
	pm.r = ARGB.r * a
	pm.g = ARGB.g * a
	pm.b = ARGB.b * a
	Return pm
End Sub

Public Sub SetHSV(a As Float, x As Int, y As Int, h As Int, s As Float, v As Float)
	Dim r, g, b As Float
	Dim hi As Int = Floor(h / 60)
	Dim hbucket As Int =  hi Mod 6
	Dim f As Float = h / 60 - hi
	Dim p As Float = v * (1 - s)
	Dim q As Float = v  * (1 - f * s)
	Dim t As Float = v * (1 - (1 - f) * s)
	Select hbucket
		Case 0
			r = v
			g = t
			b = p
		Case 1
			r = q
			g = v
			b = p
		Case 2
			r = p
			g = v
			b = t
		Case 3
			r = p
			g = q
			b = v
		Case 4
			r = t
			g = p
			b = v
		Case 5
			r = v
			g = p
			b = q
	End Select
	tempARGB.a = a
	tempARGB.r = r * 255
	tempARGB.g = g * 255
	tempARGB.b = b * 255
	SetARGB(x, y, tempARGB)
End Sub

Public Sub SetARGB(x As Int, y As Int, ARGB As ARGBColor)
	SetPremultipliedColor(x, y, ARGBToPremultipliedColor(ARGB, tempPM))
End Sub

Public Sub SetPremultipliedColor (x As Int, y As Int, Premultiplied As PremultipliedColor)
	Dim cp As Int = x * 4 + y * mWidth * 4
	buffer(cp + ri) = Premultiplied.r
	buffer(cp + gi) = Premultiplied.g
	buffer(cp + bi) = Premultiplied.b
	buffer(cp + ai) = Premultiplied.a
End Sub

Public Sub GetPremultipliedColor (x As Int, y As Int, Result As PremultipliedColor) As PremultipliedColor
	Dim cp As Int = x * 4 + y * mWidth * 4
	Result.r = buffer(cp + ri)
	Result.g = buffer(cp + gi)
	Result.b = buffer(cp + bi)
	Result.a = buffer(cp + ai)
	Return Result
End Sub

Public Sub GetColor (x As Int, y As Int) As Int
	GetARGB(x, y, tempARGB)
	Return Bit.ShiftLeft(tempARGB.a, 24) + Bit.ShiftLeft(tempARGB.r, 16) + Bit.ShiftLeft(tempARGB.g, 8) + tempARGB.b
End Sub

Public Sub GetARGB (x As Int, y As Int, Result As ARGBColor) As ARGBColor
	Dim cp As Int = x * 4 + y * mWidth * 4
	Result.a = Bit.And(0xff, buffer(cp + ai))
	Dim af As Float = Result.a / 255
	Result.r = Bit.And(0xff, buffer(cp + ri)) / af
	Result.g = Bit.And(0xff, buffer(cp + gi)) / af
	Result.b = Bit.And(0xff, buffer(cp + bi)) / af
	Return Result
End Sub

Public Sub IsTransparent(x As Int, y As Int) As Boolean
	Dim cp As Int = x * 4 + y * mWidth * 4
	Return buffer(cp + ai) = 0
End Sub

Public Sub getBitmap As B4XBitmap
	#if B4A
	Dim bf As JavaObject
	bf = bf.InitializeStatic("java.nio.ByteBuffer").RunMethodJO("wrap", Array(buffer))
	Dim jo As JavaObject = b4abmp
	jo.RunMethod("setDensity", Array(160))
	jo.RunMethod("copyPixelsFromBuffer", Array(bf))
	Return b4abmp
	#else if B4J
	Dim writer As JavaObject = WriteableImage.RunMethod("getPixelWriter", Null)
	Dim PixelFormat As JavaObject
	PixelFormat.InitializeStatic("javafx.scene.image.PixelFormat")
	writer.RunMethod("setPixels", Array(0, 0, mWidth, mHeight, PixelFormat.RunMethod("getByteBgraPreInstance", Null), _
		buffer, 0, mWidth * 4))
	Return WriteableImage
	#Else
	Dim no As NativeObject = Me
	Dim bmp As Bitmap = no.RunMethod("bufferToUIImage:::", Array(no.ArrayToNSData(buffer), mWidth, mHeight))
	Return bmp
	#end if
End Sub



#Region ************  Drawing methods ***************
Public Sub FillRect (rect As B4XRect, color As Int)
	ARGBToPremultipliedColor(ColorToARGB(color, tempARGB), tempPM)
	Dim cp As Int
	Dim right As Int = Min(mWidth - 1, rect.Right)
	For y = Max(0, rect.Top) To Min(mHeight - 1, rect.Bottom)
		Dim Start As Int = Max(0, rect.Left)
		cp = y * mWidth * 4 + Start * 4
		For x = Start To right
			SetPremultipliedColor(x, y, tempPM)
			cp = cp + 4
		Next
	Next
End Sub

Public Sub DrawBitmap (bmp As B4XBitmap, TargetRect1 As B4XRect)
	If bmp.Width <> TargetRect1.Width Or bmp.Height <> TargetRect1.Height Then
		bmp = bmp.Resize(TargetRect1.Width, TargetRect1.Height, False)
	End If
	Dim Src As BitmapCreator
	Src.Initialize(bmp.Width, bmp.Height)
	Src.CopyPixelsFromBitmap(bmp)
	Dim SrcRect As B4XRect
	SrcRect.Initialize(0, 0, Src.mWidth, Src.mHeight)
	DrawBitmapCreator(Src, SrcRect, TargetRect1.Left, TargetRect1.Top)
End Sub

Public Sub DrawBitmapCreator (Source As BitmapCreator, SrcRect As B4XRect, TargetX As Int, TargetY As Int)
	If TargetY >= mHeight Or TargetX >= mWidth Then Return
	TargetX = Max(0, TargetX)
	TargetY = Max(0, TargetY)
	Dim SrcPM As PremultipliedColor
	Dim DeltaX As Int = Min(mWidth - 1 - TargetX, Min(SrcRect.Width - 1, Source.mWidth - 1 - SrcRect.Left))
	Dim DeltaY As Int = Min(mHeight - 1 - TargetY, Min(SrcRect.Height - 1, Source.mHeight - 1 - SrcRect.Top))
	For x = TargetX To TargetX + DeltaX
		For y = TargetY To TargetY + DeltaY
			Source.GetPremultipliedColor(x - TargetX + SrcRect.Left, y - TargetY + SrcRect.Top, SrcPM)
			BlendPixel(x, y, SrcPM)
		Next
	Next
End Sub

Public Sub BlendPixel (x As Int, y As Int, SrcPM As PremultipliedColor)
	If SrcPM.a <> -1 Then
		If SrcPM.a = 0 Then Return
		SrcPM.a = Bit.And(0xff, SrcPM.a)
		GetPremultipliedColor(x, y, tempPM)
		Dim SrcAF As Float = 1 - (SrcPM.a / 255)
		SrcPM.r =  Bit.And(0xff, SrcPM.r) + SrcAF *  Bit.And(0xff, tempPM.r)
		SrcPM.g =  Bit.And(0xff, SrcPM.g) + SrcAF *  Bit.And(0xff, tempPM.g)
		SrcPM.b =  Bit.And(0xff, SrcPM.b) + SrcAF *  Bit.And(0xff, tempPM.b)
		SrcPM.a =  Bit.And(0xff, SrcPM.a) + SrcAF *  Bit.And(0xff, tempPM.a)
	End If
	SetPremultipliedColor(x, y, SrcPM)
End Sub

#End Region

#Region *****************  Native code ***********

#if OBJC
- (void)UIImageToBuffer:(UIImage*) bmp :(NSData*)data :(int)width :(int)height{
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	CGContextRef context = CGBitmapContextCreate([(NSMutableData*)data mutableBytes], 
			width, 
			height, 
			8, 
			4 * width, 
			colorSpace, 
			kCGImageAlphaPremultipliedLast); 
	if(!context) {
		NSLog(@"Failed to create bitmap context");
	}		
	CGColorSpaceRelease(colorSpace);
	CGImageRef imageRef = bmp.CGImage;
	CGRect rect = CGRectMake(0, 0, width, height);
	CGContextDrawImage(context, rect, imageRef);
	CGContextRelease(context);
}
- (UIImage*) bufferToUIImage:(NSData*)data :(int)width :(int)height {
	CGDataProviderRef provider = CGDataProviderCreateWithCFData( (CFDataRef) data );
	CGColorSpaceRef colorSpaceRef = CGColorSpaceCreateDeviceRGB();
	CGBitmapInfo bitmapInfo = kCGBitmapByteOrderDefault | kCGImageAlphaPremultipliedLast;
	CGColorRenderingIntent renderingIntent = kCGRenderingIntentDefault;
	
	CGImageRef iref = CGImageCreate(width, 
				height, 
				8, 
				32, 
				4 * width, 
				colorSpaceRef, 
				bitmapInfo, 
				provider,	
				NULL,	
				YES,
		renderingIntent);
	UIImage* image = [UIImage imageWithCGImage:iref scale:[[UIScreen mainScreen] scale] orientation:UIImageOrientationUp];
	CGColorSpaceRelease(colorSpaceRef);
	CGImageRelease(iref);
	CGDataProviderRelease(provider);
	return image;
}
#End If

#End Region